// bins.h - Binning of the spectral data
// Copyright (C) 2009, 2012, 2015 Rok Zitko

#ifndef _bins_h_
#define _bins_h_

/* Bin spectral peaks. P::bins defines the number of bins per energy
 decade. The lowest and highest energies are defined by the zero-th and
 last NRG energy scale. */

class Bins
{
 private:
   unsigned int binsperdecade;
   unsigned int nrbins;

   // log10 of the lower energy limit of the bins
   double log10emin;

 public:
   Spikes bins; // Note: Spikes is vector of (t_eigen,t_weight) pairs

   void loggrid(); 
   void loggrid_std(); // standard NRG logarithmic mesh for spec. funcions
   void loggrid_acc(); // log grid with shifted accumulation point
   
   Bins();

   inline void add(double energy, t_weight weight);
   inline void add_std(double energy, t_weight weight);
   inline void add_acc(double energy, t_weight weight);

   void add_value(double energy, t_weight weight);
   void merge(const Bins &b);
   void trim();
   t_weight total_weight();

   // Masquerade Bins objects as Spikes objects
   operator Spikes & () { return bins; }
   typedef Spikes::iterator iterator;
   typedef Spikes::const_iterator const_iterator;
   iterator begin() { return bins.begin(); }
   iterator end() { return bins.end(); }
};

Bins::Bins()
{
  // Default behavior: standard log mesh
  loggrid();
}

// These control the energy range of the bins. This is the shift in
// the (10-base) exponent of the top-most and bottom-most bins.
// WAS: 1.0 for both.
#define MAX_BIN_SHIFT 2.0
#define MIN_BIN_SHIFT 2.0

void Bins::loggrid()
{
   my_assert(P::bins > 0);
   binsperdecade = P::bins;

   if (P::accumulation > 0.0) {
      loggrid_acc();
   } else {
      loggrid_std();
   }
}

void Bins::loggrid_acc()
{
   const double a = P::accumulation;
   my_assert(a > 0.0);

   const double emax = SCALE(0)*100.0;
   const double emin = LAST_STEP_SCALE()/100.0;
   
   bins.resize(0);
   for (double e = emin; e <= emax; e *= pow(10.0, 1.0/binsperdecade)) {
      double x = (emax-a)/emax * e + a;
      bins.push_back(make_pair(x, 0.0));
   }
   
   if (P::linstep != 0.0) {
      my_assert(P::linstep > 0.0);
      for (double e = a; e > 0.0; e -= P::linstep)
	bins.push_back(make_pair(e, 0.0));
   }   
   
   bins.push_back(make_pair(DBL_MIN, 0.0)); // add zero point
   
   sort(bins.begin(), bins.end(), sortfirst());
   nrbins = bins.size();
   my_assert(nrbins >= 2);
}

void Bins::loggrid_std()
{
  // Trick: use ceil/floor to obtain uniform binning grids for different
  // values of the twist parameter z! NOTE: this will silently discard
  // spectral peaks far outside the conduction band!!
  log10emin = floor(log10(LAST_STEP_SCALE())) - MIN_BIN_SHIFT;
  double log10emax = ceil(log10(SCALE(0))) + MAX_BIN_SHIFT;

  nrbins = (unsigned int)((log10emax-log10emin)*binsperdecade+1.0);

  bins.resize(nrbins); // Note: Spikes is a vector type!

   // bins[i].first is the energy of the representative point (it is NOT an
   // interval boundary!). Thus bins[i].second represents the spectral
   // weight at energy bins[i].first. This implies that when the spectral
   // weight is collected [see function add()], the weight of a delta peak
   // with energy between two consecutive representative points is devided
   // between the two points proportionally with respect to the energies of
   // the points. This also implies that in broadening, we may consider
   // "bins" as delta peaks, rather than as interval representations (true
   // bins). In other words, what we are doing here is coarse graining
   // rather than binning!
  for (unsigned int i = 0; i < nrbins; i++)
    bins[i] = make_pair(pow(10.0, log10emin + (double)i/binsperdecade), 0.0);
}

// abs() depends on t_weight data type. It should work correctly
// for both double and complex<double>.
#ifdef NRG_REAL
inline double ABS(double x)
{
   return abs(x);
}
#endif

#ifdef NRG_COMPLEX
inline double ABS(complex<double> x)
{
   return abs(x);
}
#endif

// Unbiased assignment of the spectral weight to bins.
inline void Bins::add(double energy, t_weight weight)
{
   if (ABS(weight) < P::DISCARD_IMMEDIATELY * energy)
     return;

   if (P::accumulation > 0.0) {
      add_acc(energy, weight);
   } else {
      add_std(energy, weight);
   }
}

inline void Bins::add_std(double energy, t_weight weight)
{
   const double lg = log10(energy);
   const double pos = (lg-log10emin)*binsperdecade;

   const int index = int(pos);

   // Important: if 'energy' is lower than the lower limit of the
   // first interval, the weight is assigned to the first bin. This
   // is especially relevant for collecting the omega=0 data in
   // bosonic correlators. (rz, 25 Oct 2012)
   if (index < 0) {
      bins[0].second += weight;
      return;
   } else if ((unsigned int)(index) >= nrbins-1) {
      bins[nrbins-1].second += weight;
      return;
   }

   const double rem = pos-index;
   bins[index].second += (1.0-rem) * weight;
   bins[index+1].second += rem * weight;
}

inline void Bins::add_acc(double energy, t_weight weight)
{
   const unsigned int len = bins.size();

   for (unsigned int i = 0; i < len-1; i++) {
      const double e1 = bins[i].first;
      const double e2 = bins[i+1].first;
      my_assert (e1 < e2);
      if (e1 < energy && energy < e2) {
	 const double dx = e2-e1;
	 const double reldist = (energy-e1)/dx;
	 bins[i].second += (1.0-reldist) * weight;
	 bins[i+1].second += reldist * weight;
	 return;
      }
   }
   // Note: if no suitable interval is found, the weight is discarded!
}

// IMPORTANT: use add_value() instead of add() for any type of data
// except the standard log grid. Here the bin is chosen by exact
// numerical comparison with the value of 'energy'.
inline void Bins::add_value(double energy, t_weight weight)
{
   const unsigned int len = bins.size();
   
   for (unsigned int i = 0; i < len; i++) {
      if (bins[i].first == energy) {
	 bins[i].second += weight;
	 return;
      }
   }
   // or else 
   bins.push_back(make_pair(energy, weight));
}

// Merge two bins. They need to agree in the representative energies
// (first element of the pairs).
void Bins::merge(const Bins &b)
{
   my_assert(bins.size() == b.bins.size());
   my_assert(bins.size() == nrbins);

   for (unsigned int i = 0; i < nrbins; i++) {
      my_assert(bins[i].first == b.bins[i].first);
      bins[i].second += b.bins[i].second;
   }
}

t_weight Bins::total_weight()
{
   t_weight sum = 0;
   for (unsigned int i = 0; i < nrbins; i++) {
      sum += bins[i].second;
   }
   return sum;
}

// Only keep bins which are "heavy" enough.
void Bins::trim()
{
   Spikes bins2;
   bucket discarded_weight_abs;
   unsigned int nr = bins.size();
   bins2.reserve(nr);
   // nr-1, because we need to compute the energy interval size 'ewidth'
   for (unsigned int i = 0; i < nr-1; i++) {
      const double e = bins[i].first;
      const t_weight w = bins[i].second;
      const double enext = bins[i+1].first; // increasing!
      my_assert(enext > e);
      const double ewidth = enext-e;

      if (ABS(w) < P::DISCARD_TRIM * ewidth) {
         discarded_weight_abs += abs(w);
      } else {
         bins2.push_back(bins[i]);
      }
   }
   // Always keep the last one.. This ensures that we keep
   // information about the energy range on which the calculations
   // has been performed. [XXX: should we also keep the first one?]
   bins2.push_back(bins[nr-1]);
   bins.swap(bins2);

   if (logletter('T'))
     nrgdump(discarded_weight_abs) << endl;

   if (discarded_weight_abs > 1e-6) {
      cerr << "WARNING: we are probably discarding too much weight!" << endl;
   }
}

#endif // _bins_h_
