// Numerical renormalization group
// (c) Rok Zitko, rok.zitko@ijs.si, 2005-2015

#ifndef _nrg_h_
#define _nrg_h_

#include <iostream>
#include <iomanip>
#include <utility>
#include <boost/serialization/map.hpp>
#include <boost/serialization/list.hpp>
#include <deque>
#include <fstream>
#include <sstream>
#include <set>
#include <numeric>
#include <boost/serialization/vector.hpp>
#include <algorithm>
#include <cassert>
#include <cmath>
#include <cfloat>
#include <climits>
#include <cstring>
#include <complex>
#include <boost/serialization/complex.hpp>
#include <functional>

#include "portabil.h"

#include <unistd.h>

// If we have gmp, we can do the tridiagonalisation.
#ifdef HAVE_LIBGMP
#include <gmp.h>
#define TRIDIAG
#endif

#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/symmetric.hpp>
#include <boost/numeric/ublas/operation.hpp>
#include <boost/static_assert.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/foreach.hpp>
#include <boost/io/ios_state.hpp>
#include <boost/tuple/tuple.hpp>
#include <boost/array.hpp>
#include <boost/shared_ptr.hpp>

#include <boost/unordered_map.hpp>
//#include <tr1/unordered_map> // TR1 hashed map container
//#include <unordered_map> // Now part of C++11 standard
#define MY_UNORDERED_MAP boost::unordered_map

#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>

#include <boost/archive/binary_iarchive.hpp>
#include <boost/archive/binary_oarchive.hpp>

#include <boost/serialization/base_object.hpp>
#include <boost/serialization/utility.hpp>
#include <boost/serialization/list.hpp>

#define FOREACH BOOST_FOREACH

using namespace std;
using namespace boost::numeric;
using namespace boost::numeric::ublas;
using namespace boost::tuples;
using namespace boost::lambda;

// Numeric bindings to BLAS/LAPACK (ATLAS/MKL)

#include <boost/numeric/bindings/traits/ublas_vector.hpp>
#include <boost/numeric/bindings/traits/ublas_matrix.hpp>
#include <boost/numeric/bindings/atlas/cblas.hpp>

using namespace boost::numeric::bindings::atlas;
namespace atlas = boost::numeric::bindings::atlas;

// This must be set manually through CXXFLAG=-DNRG_MPI
#ifdef NRG_MPI

#include <boost/optional.hpp>
#include <boost/mpi/environment.hpp>
#include <boost/mpi/communicator.hpp>
#include <boost/mpi/collectives.hpp>
namespace mpi = boost::mpi;

mpi::environment *mpienv;
mpi::communicator *mpiw;
int myrank;

const int TAG_HELLO = 1;
const int TAG_EXIT = 2;
const int TAG_DIAG = 3;
const int TAG_SYNC = 4;
const int TAG_MATRIX = 5;
const int TAG_INVAR = 6;
const int TAG_EIGEN = 7;

const int TAG_MATRIX_SIZE = 8;
const int TAG_MATRIX_LINE = 9;

const int TAG_EIGEN_INT = 10;
const int TAG_EIGEN_VEC = 11;
const int TAG_EIGEN_RMAXVALS = 12;

#endif // NRG_MPI

// Support for compiler dependant optimizations

#ifdef __GNUC__
#define PUREFNC __attribute__ ((pure))
#define CONSTFNC __attribute__ ((const))
#else
#define PUREFNC
#define CONSTFNC
#endif

// Hard coded & default filenames

// Spectral densities and temperature dependant quantities
#define FN_ENERGIES_NRG "energies.nrg"
#define FN_ENERGIES_DMNRG "energies.dmnrg"
#define FN_ANNOTATED "annotated.dat"

#define FN_TD "td"
#define FN_CUSTOM "custom"
#define FN_CUSTOMFDM "customfdm"
#define FN_CUSTOMSQ "customsq"

// Input files
#define FN_DATA "data"
#define FN_PARAM "param"

// Density-matrix approach
#define FN_UNITARY "unitary"
#define FN_RHO "rho"
#define FN_RHOFDM "rhofdm"

#define FN_SUBSPACES "subspaces.dat"

// Directory for temporary files. No trailing slash.
#define DIR_UNITARY "."

#define T_WIDTH 12

// Warning: not thread safe!
class Timing {
private:
  double all, timer;
  bool running;
  map<string, double> t;
  bool reportonexit;

public:  
  Timing(bool _reportonexit = true) :
   reportonexit(_reportonexit) {

    all = gettime();
    timer = gettime();
    running = false;
  }

  void start(void) {
    my_assert(!running);
    running = true;
    timer = gettime();
  }

  double stop(void) {
    my_assert(running);
    running = false;
    double end = gettime();
    return end-timer;
  }

  void add(string timer) {
    if (!t.count(timer)) 
       t[timer] = 0.0; // initialize
    t[timer] += stop();
  }

  double value(string timer) const {
    if (!t.count(timer)) 
       return 0.0;
    return t.find(timer)->second;
  }

  double total() {
     double end = gettime();
     return end-all;
  }   
   
  void report(void) {
    const double tall = total();

    cout << endl;
    cout << "Timing report";
#ifdef NRG_MPI
    cout << " (process " << myrank << ")" << endl;
#else
    cout << endl;
#endif
    cout << "=============" << endl;
    cout << setw(T_WIDTH) << "All" << ": " << long(tall) << " s" << endl;
    double tsum = 0.0;
    map<string, double>::iterator i;
    for (i = t.begin(); i != t.end(); i++) {
      // Only show those that contribute more than 1% of the total time!
      if ( (i->second/tall) > 0.01) {
	 cout << setw(T_WIDTH) << i->first << ": " 
	      << long(i->second) << " s" << endl;
	 if (i->first[0] != '*') 
	   tsum += i->second;
      }
    }
    cout << setw(T_WIDTH) << "Other" << ": " << long(tall-tsum) << " s" << endl;
  }

  ~Timing() {
    if (reportonexit)
      report();
  }
};

// Higher-level timing code

// Time a section for as long as the object is in scope.
class TimeScope 
{
 private:
   Timing & timer;
   string timer_name;
   
 public:
   TimeScope(Timing & _timer, const string &_timer_name) :
    timer(_timer), timer_name(_timer_name) {
       timer.start();
    }
   
   ~TimeScope() {
      timer.add(timer_name);
   }
};

#define TIME(timer_name) TimeScope timer(t, timer_name)
#define TIME_SECTION(timer_name, section)  { TIME(timer_name); section }

// Stores maximal memory usage at various breakpoints.
// This is useful for estimating memory requirements at various
// points of the execution path.

#define MS_WIDTH 12

class MemoryStats {
private:
  map<string, int> maxvals;
  bool reportonexit;
  int peakusage;

public:
  MemoryStats(bool rpexit = true) {
    reportonexit = rpexit;
    peakusage = 0;
  }

  // Intermediate level routine.
  int used() {
     const int memused = memoryused();
     peakusage = max(peakusage, memused);
     return memused; 
  }
	
  // Sample memory usage at an arbitrarily named "breakpoint".
  int check(string breakpoint) {
    const int memused = used();

    if (maxvals.count(breakpoint) == 0) 
      maxvals[breakpoint] = 0;
     
    maxvals[breakpoint] = max(maxvals[breakpoint], memused);
    return memused;
  }
    
  void report(ostream &F = cout) const {
#ifdef HAS_MEMORY_USAGE     
     
    F << endl;
    F << "Memory usage report";
#ifdef NRG_MPI
    F << " (process " << myrank << ")" << endl;
#else
    F << endl;
#endif
    F << "===================" << endl;

    int topusage = 0; // top usage recorded by check()
    for (map<string, int>::const_iterator i = maxvals.begin(); i != maxvals.end(); i++) {
       topusage = max(topusage, i->second);
    }
    if (topusage != 0) {
       for (map<string, int>::const_iterator i = maxvals.begin(); i != maxvals.end(); i++) {
	  F << setw(MS_WIDTH) << i->first << ": " << i->second << " kB" << endl;
       }
    }
    my_assert(topusage <= peakusage);
    F << endl << "Peak usage: " << peakusage << " kB" << endl;
#endif
  }   

  ~MemoryStats() {
     if (reportonexit)
       report();
  }
};

#define HIGHPREC(val) setw(30) << setprecision(16) << (val) \
                                 << setprecision(COUT_PRECISION)  

#endif
