void calc_generic_FT(const Eigen & diagIp, const Eigen & diagI1, 
		     const Matrix & op1II, const Matrix & op2II,
		     double sign, t_factor spinfactor,
		     ChainSpectrum * cs)
{
   using namespace STAT;
   const bool boolT0opt = P::T0opt >= 0; // Enable T=0 optimizations
   const unsigned int T0opt = (boolT0opt ? P::T0opt : 0); // index limit in for loops over states

   const unsigned int dimp = diagIp.getnr();
   const unsigned int dim1 = diagI1.getnr();

   for (unsigned int r1 = 0; r1 < dim1; r1++) {
      const t_eigen E1 = diagI1.value(r1);
          
      for (unsigned int rp = 0; rp < dimp; rp++) {
	 if (boolT0opt && r1 > T0opt && rp > T0opt) // does not contribute at T=0
	   continue;
	   
	 const t_eigen Ep = diagIp.value(rp);

	 DELTA d;
	 // The first matrix element is conjugated!
	 d.weight = (spinfactor/Zft) * CONJ_ME(op1II(r1, rp)) * op2II(r1, rp)
	   * ( (-sign) * exp(-E1 * scT) + exp(-Ep * scT) ); // sign!
	 d.energy = E1 - Ep;

	 cs->add(scale * d.energy, d.weight);
      } // loop over r1
   } // loop over rp
}

// Calculation of the temperature-dependent linear conductrance G(T) using
// the linear response theory & impurity-level spectral density. Note that
// we are not calculating a spectral function (i.e. a collection of delta
// peaks), but rather a tabulated G(T), so binning needs to be turned off.
// See Yoshida, Seridonio, Oliveira, arxiv:0906.4289, Eq. (8).
void calc_generic_GT(const Eigen & diagIp, const Eigen & diagI1, 
		     const Matrix & op1II, const Matrix & op2II,
		     double sign, t_factor spinfactor,
		     ChainSpectrum * cs,
		     const Invar &Ip, const Invar &I1, int power)
{
   my_assert(sign == S_FERMIONIC); // restricted implementation
   
   const double temperature = P::gtp * STAT::scale; // in absolute units!
   const double beta = 1.0/temperature; 

   const unsigned int dimp = diagIp.getnr();
   const unsigned int dim1 = diagI1.getnr();

   weight_bucket value;

   for (unsigned int r1 = 0; r1 < dim1; r1++) {
      const t_eigen E1 = diagI1.value(r1);
          
      for (unsigned int rp = 0; rp < dimp; rp++) {
	 const t_eigen Ep = diagIp.value(rp);
 	 // Note that Zgt needs to be calculated with the same
 	 // 'temperature' parameter that we use for the exponential
 	 // functions in the following equation.
	 value += beta * (spinfactor/STAT::Zgt) * CONJ_ME( op1II(r1, rp) ) * op2II(r1, rp) 
	   / (exp(+E1 * STAT::scale * beta) + exp(+Ep * STAT::scale * beta)) 
	     * pow((E1-Ep)*STAT::scale, power);
      } // loop over r1
   } // loop over rp

   cs->add(temperature, value);
}

const double CHIT_WEIGHT_TOL = 1e-8; // where to switch to simple form

// weight=(exp(-beta Em)-exp(-beta En))/(beta En-beta Em).
// NOTE: arguments En, Em are order omega_N, while beta is order
// 1/omega_N, thus the combinations betaEn and betaEm are order 1.
// Also En>0, Em>0, since these are excitation energies !
inline t_weight chit_weight(double En, double Em, double beta)
{
   const double betaEn = beta * En;
   const double betaEm = beta * Em;
   const double x = betaEn-betaEm;
   
   // see chit_calculation_tests.nb
   if (abs(x) > CHIT_WEIGHT_TOL) {
      // If one of {betaEm,betaEn} is small, one of exp() will have a
      // value around 1, the other around 0, thus the overall result
      // will be approximately +-1/x.
      return (exp(-betaEm)-exp(-betaEn))/x;
   } else {
      // Special case for Em~En. In this case, we are integrating a
      // constant over tau\in{0,\beta}, and dividing this by beta we
      // get 1. What remains is the Boltzmann weight exp(-betaEm).
      return exp(-betaEm);
   }
}
 
// Calculation of the temperature-dependent susceptibility chi_AB(T)
// using the linear response theory and the matrix elements of global
// operators. Binning needs to be turned off.
void calc_generic_CHIT(const Eigen & diagIp, const Eigen & diagI1, 
                       const Matrix & op1II, const Matrix & op2II,
                       double sign, t_factor spinfactor,
                       ChainSpectrum * cs,
                       const Invar &Ip, const Invar &I1)
{
   // restricted implementation: singlet operators only!
   my_assert(sign == S_BOSONIC);
   const unsigned int dimp = diagIp.getnr();
   const unsigned int dim1 = diagI1.getnr();
   my_assert(dim1 == dimp);
   
   const double temperature = P::chitp * STAT::scale; // in absolute units!
   const double beta = 1.0/temperature; 

   
   weight_bucket w;
   
   for (unsigned int r1 = 0; r1 < dim1; r1++) {
      const t_eigen E1 = diagI1.value(r1);
          
      for (unsigned int rp = 0; rp < dimp; rp++) {
	 const t_eigen Ep = diagIp.value(rp);
 	 // Note that Zchit needs to be calculated with the same
 	 // 'temperature' parameter that we use for the exponential
 	 // functions in the following equation.
	 // NOTE: E1, Ep are quantities of order 1, while STAT::scale
	 // is the current omega_N scale. Thus these are absolute
	 // energies!
	 const t_weight weight = chit_weight(STAT::scale * E1, 
					     STAT::scale * Ep,
					     beta);
	 assert_isfinite(weight);

	 w += op1II(r1, rp) * op2II(rp, r1) * weight;

         // NOTE: there is no beta factor here. This implies that the
         // output is actually chi/beta = k_B T chi. Which is actually what
         // we prefer anyway..
      } // loop over r1
   } // loop over rp

   const t_weight result = spinfactor/STAT::Zchit * t_weight(w);
   cs->add(temperature, result);
}
